/*  Filename: BrowserOverlay.js
 * $Revision: 1.25 $
 *     $Date: 2008/05/23 21:29:50 $
 *
 * TODO: description
 * 
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TabLang extension for Mozilla Firefox.
 *
 * The Initial Developer of the Original Code is Andrey Lokot.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrey Lokot <crazy-eye@yandex.ru>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// Global objects. Are the same for all browser.xul windows
var tablangMain = undefined;
var tablangOptions = undefined;

// Wrapper for TabLangSwitcher component. Object operates with variables of informal type LANG
// which can be both LANG_TAG and LANG_ID so it makes translation between LANG_TAG and LANG_ID.
// transparent for the caller. "switcher" member can be used to access TabLangSwitcher component directly.
function tablangSwitcherWrapper() {
	// "switcher" variable is used to access TabLangSwitcher component (through ITabLangSwitcher interface)
	this.switcher = Components.classes["tablang@mozdev.org/tablangswitcher;1"]
									.createInstance()
									.QueryInterface(Components.interfaces.ITabLangSwitcher);
	// "currentLang" variable of type LANG which is used to get and set current system input language
	// "currentLang" getter returns current system LANG. It uses
	// GetCurrentLangId() so actual return value is LANG_ID.
	this.__defineGetter__("currentLang", function() {
		try {
			return this.switcher.GetCurrentLangId();
		}
		catch (err) {
			var errorText = "[TabLang] Exception in currentLang getter\n" + err;
			Components.utils.reportError(errorText);
			alert(errorText);
			// Trying to hide failure
			return tablangOptions.defaultOverride;
		}
	});
	// "currentLang" setter takes "lang" parameter of type LANG. Uses SetCurrentLangId() 
	// if "lang" is LANG_ID or SetCurrentLangTag() if "lang" is LANG_TAG
	this.__defineSetter__("currentLang", function(lang) {
		if (lang == undefined) {
			// Possibly due to bug 366705 (https://bugzilla.mozilla.org/show_bug.cgi?id=366705):
			// TabOpen may not fire when user tries to load nonexisting page from the local drive
			// so browser.tablangLang would not be set for this tab. This was fixed in Firefox 3.0b1
			var errorText = "[TabLang] Error: language not defined for this tab";
			Components.utils.reportError(errorText);
			alert(errorText);
			lang = tablangOptions.defaultLang;
		}
		try {
			if (typeof(lang) == "number") {
				this.switcher.SetCurrentLangId(lang);
			}
			else {
				this.switcher.SetCurrentLangTag(lang);
			}
		}
		catch (err) {
			var errorText = "[TabLang] Exception in currentLang setter\nnewlang=" + lang + "\n" + err;
			Components.utils.reportError(errorText);
			alert(errorText);
			// New language wasn't set but we still can try to continue execution
		}
	});
	// forceLangTag() function converts LANG to LANG_TAG
	// Takes "lang" parameter of type LANG and returns corresponding LANG_TAG
	// If "lang: is actually LANG_TAG returns "lang" itself. Otherwise
	// uses LangTagFromId() to convert LANG_ID to LANG_TAG.
	this.forceLangTag = function(lang) {
		try {
			return (typeof(lang) == "number") ? this.switcher.LangTagFromId(lang) : lang;
		}
		catch (err) {
			var errorText = "[TabLang] Exception in forceLangTag()\nlang=" + lang + "\n" + err;
			Components.utils.reportError(errorText);
			alert(errorText);
			// Seems to be the error in caller code so throwing exception out
			throw err;
		}
	};
}

// tablangState object is attached to current browser state to new tab
// It is used later to set language for the new tab.
function tablangState(browser, uri, lang) {
	this.browser = browser;
	this.uri = uri;
	this.lang = lang;
}

// Extension initialization function
// Is called for every new "browser.xul" window.
var tablangOnLoadMainWindow = function(event) {
	// removing event listener as it will never be called again
	window.removeEventListener("load", tablangOnLoadMainWindow, false);
	// Searching for already existing tablangMain component (in another browser window)
	var enumerator = Components.classes["@mozilla.org/appshell/window-mediator;1"]
						.getService(Components.interfaces.nsIWindowMediator)
						.getEnumerator("navigator:browser");
	var found = false;
	var winFound;
	while (enumerator.hasMoreElements()) {
		winFound = enumerator.getNext();
		if (winFound.tablangMain) {
			found = true;
			break;
		}
	}
	if (found) {
		// If extension was already initialized in another window
		// creats links to existing objects
		tablangOptions = winFound.tablangOptions;
		tablangMain = winFound.tablangMain;
		// As TabOpen event is not send for the first tab in the window
		// some initialization for it should be performed
		var browser = gBrowser.getBrowserAtIndex(0);
		tablangMain.onNewTab(window, browser);
		if (Components.classes["@mozilla.org/appshell/window-mediator;1"]
								.getService(Components.interfaces.nsIWindowMediator)
								.getMostRecentWindow(null)
									== window) {
			// If current window is focused make it internally active
			if (tablangMain.currentObject) {
				tablangMain.currentObject.tablangLang = tablangMain.wrapper.currentLang;
			}
			tablangMain.currentObject = tablangMain.lastBrowser = browser;
			tablangMain.wrapper.currentLang = tablangOptions.defaultLang;
		}
	}
	else {
		// If it's the first window perform initialization
		try {
			// tablangMain is the main extension object. Contains all the functions to work with tabs
			tablangMain = {
				// wrapper - tablangSwitcherWrapper instance
				wrapper: new tablangSwitcherWrapper,
				// currentObject - currently focused object(at least one extension supports).
				// Current system input language is implied to be the language of this object
				currentObject: gBrowser.selectedBrowser,
				// lastBrowser - last active(focused) browser: the last object
				// of "browser" type that was(or is) saved in currentObject
				// May be not focused. Is used to find "parent" for new tab
				lastBrowser: gBrowser.selectedBrowser,
				// urlbarLang - the language of urlbars of all windows. Is the same as tablangLang variable 
				// for urlbar element Will be set to tablangOptions.urlbarLang after tablangOptions initialization
				urlbarLang: undefined,
				// For each tab tablangLang variable is added to browser object to save language for this tab
				// The following functions are event handlers. They may take these input parameters:
				//   win     - current windows (is needed as there is only one instance of the object)
				//   browser - browser object for which the event was sent
				// onSelectTab() is called when on of the tabs receives focus
				onSelectTab: function(win) {
					// Checking if focus was actually received by 
					// some tab but not the active one (if any)
					if (this.currentObject != win.gBrowser.selectedBrowser) {
						if (this.currentObject) {
							// If there was previous focused object saving its language
							this.currentObject.tablangLang = this.wrapper.currentLang;
						}
						this.currentObject = this.lastBrowser = win.gBrowser.selectedBrowser;
						// Loading saved language for selected tab
						this.wrapper.currentLang = this.currentObject.tablangLang;
					}
				},
				// onNewTab() is called when new tab is created. 
				// Initializes new tab and attaches tablangState to it
				onNewTab: function(win, browser) {
					browser.tablangLang = tablangOptions.defaultLang;
					if (this.lastBrowser) {
						browser.tablangSavedState = new tablangState(
							this.lastBrowser,
							this.lastBrowser.currentURI, 
							(this.lastBrowser == this.currentObject) ? 
								this.wrapper.currentLang : this.lastBrowser.tablangLang
						);
					}
				},
				// onPageLoad() is called after DOM structure for the page is created
				// so it can use document.referrer to try to find tab from which
				// current page was loaded (parent window)
				onPageLoad: function(win, browser) {
					if (browser.tablangSavedState) {
						// done is set to true if parent page was found
						var done = false;
						if (browser.contentWindow.opener) {
							// If window.opener is defined use it as parent window
							var openerBrowser = null;
							var winEnum = Components.classes["@mozilla.org/appshell/window-mediator;1"]
												.getService(Components.interfaces.nsIWindowMediator)
												.getEnumerator("navigator:browser");
							while (!openerBrowser && winEnum.hasMoreElements()) {
								var enumWindow = winEnum.getNext();
								if (enumWindow.gBrowser) {
									// Fix for unknown bug after Mouse Gestures mgB_LinksInWindows:
									// winEnum.hasMoreElements() may return not fully loaded
									// "navigator:browser" window with window.gBrowser == null
									// TODO: find the source of the problem
									openerBrowser = enumWindow.gBrowser.getBrowserForDocument(
										browser.contentWindow.opener.top.document);
								}
							}
							if (openerBrowser) {
								if (openerBrowser == browser.tablangSavedState.browser) {
									// If window.opener is the browser saved in tablangSavedState
									// using the language it had when new tab was created
									browser.tablangLang = browser.tablangSavedState.lang;
								}
								else {
									// Otherwise use current state of window.opener
									browser.tablangLang = (this.currentObject == openerBrowser) ?
										this.switcher.currentLang : openerBrowser.tablangLang;
								}
								done = true;
							}
						}
						if (!done) {
							// Trying to find parent window using document.referrer
							if (browser.contentDocument.referrer) {
								var referrerURI = Components.classes["@mozilla.org/network/io-service;1"]
												.getService()
												.QueryInterface(Components.interfaces.nsIIOService)
												.newURI(browser.contentDocument.referrer, null, null);
								if (browser.tablangSavedState.uri.equals(referrerURI)) {
									browser.tablangLang = browser.tablangSavedState.lang;
									done = true;
								}
							}
						}
						// If browser  for which we find language is focused also set system language
						if (done && browser == this.currentObject) {
							this.wrapper.currentLang = browser.tablangLang;
						}
						// Delete info saved when tab was opened as it is useless now
						delete browser.tablangSavedState;
					}
				},
				// onUrlbarFocus() is called when user begins using address
				// bar. It switches current language to urlbarLang
				onUrlbarFocus: function(win) {
					if (this.currentObject) {
						this.currentObject.tablangLang = this.wrapper.currentLang;
					}
					this.currentObject = win.document.getElementById("urlbar");
					this.wrapper.currentLang = this.currentObject.tablangLang;
				},
				// onUrlbarBlur() is called when user leaves address bar. It saves
				// urlbarLang as it should not be affected by the language user
				// may use in another controls that are not supported
				onUrlbarBlur: function(win) {
					var urlbar = win.document.getElementById("urlbar");
					if (this.currentObject == urlbar) {
						urlbar.tablangLang = this.wrapper.currentLang;
						this.currentObject = null;
					}
				}
			};
			
			// tablangOptions is used to interact with extension properties
			tablangOptions = {
				// switcher - link to TabLangSwitcher that is used in Options window
				switcher: tablangMain.wrapper.switcher,
				// prefs - preferences service instance for extension(!) properties
				prefs: Components.classes["@mozilla.org/preferences-service;1"]
							.getService(Components.interfaces.nsIPrefService)
							.getBranch("extensions.tablang."),
				// langList - list of languages available in the running system
				langList: [],
				// defaultOverride - language used instead of one in preferences if that is not available
				defaultOverride: null,
				// validated() - replaces the language if it is not availabe with defaultOverride
				// Takes "lang" - LANG_TAG to validate. Returns "lang"
				// if it is availabe or defaultOverride otherwise
				validated: function(lang) {
					return (this.langList.indexOf(lang) != -1) ? lang : this.defaultOverride;
				},
				// The following properties are correct(!) LNAG_TAGs of corresponding preferences:
				// defaultLang - the language to be used for new tabs (with no parent)
				get defaultLang() { return this.validated(this.prefs.getCharPref("defaultLang")); },
				// urlbarLang - the language to be used when editing URLs in address bar
				get urlbarLang() { return this.validated(this.prefs.getCharPref("urlbarLang")); }
			};
			// Loading langList
			var list = { value: [] };
			var count = { value: list.value.count };
			tablangMain.wrapper.switcher.ListLangTags(count, list);
			tablangOptions.langList = list.value;
			// Setting defaultOverride to defaultLang if it is correct or
			// to the first language in the list otherwise
			tablangOptions.defaultOverride = tablangOptions.defaultLang;
			if (tablangOptions.defaultOverride == null) {
				tablangOptions.defaultOverride = tablangOptions.langList[0];
			}
			// Initializing urlbarLang using just loaded tablangOptions
			tablangMain.urlbarLang = tablangOptions.urlbarLang;
			
			// As TabOpen event is not send for the first tab in the window
			// some initialization for it should be performed
			tablangMain.currentObject.tablangLang = tablangOptions.defaultLang;
			tablangMain.wrapper.currentLang = tablangOptions.defaultLang;
			
		}
		catch (err) {
			// Can't initialize extesnion. Erasing variables as they are wrong and
			// should not be found from another browser windows
			tablangMain = undefined;
			tablangOptions = undefined;
			var errorText = "[TabLang] Exception during window initialization\n" + err;
			Components.utils.reportError(errorText);
			alert(errorText);
			return; 
		}
	}
	// Initializing tablangLang variable for urlbar of this window 
	// It is the same as tablangMain.urlbarLang so it is equal for all windows
	var urlbar = document.getElementById("urlbar");
	urlbar.__defineGetter__("tablangLang", function() { return tablangMain.urlbarLang; } );
	urlbar.__defineSetter__("tablangLang", function(newlang) { tablangMain.urlbarLang = newlang; } );
	// Setting event handlers
	if (tablangMain) {
		gBrowser.addEventListener(
			"focus",
			// onSelectTab() is called every time any element inside tabbrowser
			// gets focus as focus events for browser objects are sometimes not sent
			function(event) { tablangMain.onSelectTab(window); },
			true);
		gBrowser.tabContainer.addEventListener(
			"select",
			// Switching language also when switching between tabs when tabbar is focused
			function(event) { tablangMain.onSelectTab(window); },
			false);
		gBrowser.tabContainer.addEventListener(
			"TabOpen",
			function(event) { tablangMain.onNewTab(window, event.target.linkedBrowser); },
			false);
		document.getElementById("appcontent").addEventListener(
			"DOMContentLoaded",
			function(event) {
				var doc = event.originalTarget;
				// Calling onPageLoad() only for top documents (not frames/iframes)
				if (doc.nodeName == "#document" && doc.defaultView.top == doc.defaultView) {
					tablangMain.onPageLoad(window, window.gBrowser.getBrowserForDocument(doc));
				}
			},
			true);
		gBrowser.tabContainer.addEventListener(
			"TabClose",
			function(event) {
				// Preventing variables from pointing to incorrect browser object
				if (tablangMain.lastBrowser == event.target.linkedBrowser) {
					if (tablangMain.currentObject == tablangMain.lastBrowser) {
						tablangMain.currentObject = null;
					}
					tablangMain.lastBrowser = null;
				}
			},
			false);
		window.addEventListener(
			"close",
			function(event) {
				// Preventing variables from pointing to incorrect browser object
				if (tablangMain.lastBrowser && 
						gBrowser.getBrowserForDocument(tablangMain.lastBrowser.contentDocument)) {
					if (tablangMain.currentObject == tablangMain.lastBrowser) {
						tablangMain.currentObject = null;
					}
					tablangMain.lastBrowser = null;
				}
			},
			false);
		urlbar.inputField.addEventListener(
			"focus",
			function(event) { tablangMain.onUrlbarFocus(window); },
			false);
		urlbar.inputField.addEventListener(
			"blur",
			function(event) { tablangMain.onUrlbarBlur(window); },
			false);
	}
}

window.addEventListener("load", tablangOnLoadMainWindow, false);
